#import <Cocoa/Cocoa.h>

#define XeeNoMode 0
#define XeeMoveMode 1
#define XeeCopyMode 2

@class XeeImage,XeeView,XeeCollisionPanel,XeeRenamePanel;

@interface XeeController:NSObject
{
	NSString *directory;
	NSMutableArray *dircontents;

	NSLock *loadlock;
	BOOL loader_running;
	BOOL exiting;

	XeeImage *previmage,*currimage,*nextimage;
	int previndex,currindex,nextindex;

	XeeImage *loadingimage;

	int drawer_mode,pending_mode;

	IBOutlet NSWindow *window;
    IBOutlet XeeView *imageview;
    IBOutlet NSTextField *statusbar;
    IBOutlet NSDrawer *drawer;
    IBOutlet NSTextField *drawertext;
	IBOutlet NSTableView *table;
	IBOutlet XeeCollisionPanel *collisionpanel;
	IBOutlet XeeRenamePanel *renamepanel;
}

+(NSArray *)controllers;

-(BOOL)loadImage:(NSString *)filename;
-(void)displayImage:(int)index next:(int)next;
-(void)reloadImage;
-(void)updateView;
-(void)imageLoader:(NSWindow *)win;
-(void)launchLoader;
-(void)updateImage;
-(void)updateStatusBar;

-(void)sortFiles;
-(int)findFile:(NSString *)filename;
-(int)findNextFile:(NSString *)filename;
-(void)removeFile:(NSString *)filename;
-(void)insertFile:(NSString *)filename;

-(NSWindow *)window;
-(NSDrawer *)drawer;
-(NSString *)directory;
-(NSString *)currentFilename;

-(IBAction)skipNext:(id)obj;
-(IBAction)skipPrev:(id)obj;
-(IBAction)skipFirst:(id)obj;
-(IBAction)skipLast:(id)obj;
-(void)skipToFile:(NSString *)filename;
-(void)skip:(int)step;

-(IBAction)frameSkipNext:(id)obj;
-(IBAction)frameSkipPrev:(id)obj;
-(IBAction)animationToggle:(id)obj;

-(IBAction)zoomIn:(id)sender;
-(IBAction)zoomOut:(id)sender;
-(IBAction)zoomActual:(id)sender;
-(IBAction)zoomFit:(id)sender;

-(IBAction)revealInFinder:(id)sender;

-(IBAction)renameFileFromMenu:(id)sender;
-(void)renameFile:(NSString *)filename to:(NSString *)newname;

-(IBAction)deleteFileFromMenu:(id)sender;
-(IBAction)askAndDelete:(id)sender;
-(void)deleteAlertEnd:(NSAlert *)alert returnCode:(int)res contextInfo:(NSString *)filename;
-(void)deleteFile:(NSString *)filename;

-(IBAction)moveFile:(id)sender;
-(IBAction)copyFile:(id)sender;
-(void)triggerDrawer:(int)mode;
-(void)drawerDidClose:(NSNotification *)notification;
-(void)destinationListClick:(id)sender;
-(void)destinationPanelEnd:(NSOpenPanel *)panel returnCode:(int)res contextInfo:(NSString *)filename;
-(void)attemptToTransferFile:(NSString *)filename to:(NSString *)destination mode:(int)mode;
-(void)transferFile:(NSString *)filename to:(NSString *)destination mode:(int)mode;

-(IBAction)jpegAutoRotate:(id)sender;
-(IBAction)jpegRotateCW:(id)sender;
-(IBAction)jpegRotateCCW:(id)sender;
-(IBAction)jpegRotate180:(id)sender;
-(IBAction)jpegFlipHorizontal:(id)sender;
-(IBAction)jpegFlipVertical:(id)sender;

-(void)jpegTransform:(NSString *)option;

-(IBAction)launchAppFromMenu:(id)sender;
-(IBAction)launchDefaultEditor:(id)sender;

-(void)errorMessage:(NSString *)title text:(NSString *)text;

@end



@interface XeeWindow:NSWindow
{
	IBOutlet XeeController *controller;
}

-(void)keyDown:(NSEvent *)event;
-(void)scrollWheel:(NSEvent *)event;

@end
