#import <Cocoa/Cocoa.h>
#import <OpenGL/GL.h>
#import <OpenGL/GLu.h>
#import <QuickTime/QuickTimeComponents.h>

@class XeeView,XeeImageFrame,XeeImageTile,XeeImageLoader;

@interface XeeImage:NSObject
{
	NSString *filename;
	NSDictionary *attrs;

	NSString *format;
	int width,height,depth;
	BOOL transparent;
	NSColor *back;

	BOOL stop;

	XeeView *view;
}

+(XeeImage *)imageForFilename:(NSString *)filename;
+(XeeImage *)imageForFilename:(NSString *)filename view:(XeeView *)targetview;
+(void)registerImageClass:(Class)class;
+(NSArray *)fileTypes;

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes view:(XeeView *)targetview;

-(BOOL)load;
-(BOOL)loaded;
-(void)stopLoading;
-(BOOL)hasBeenStopped;

-(int)frames;
-(void)setFrame:(int)frame;
-(int)frame;

-(BOOL)animated;
-(void)setAnimating:(BOOL)animating;
-(BOOL)animating;

//-(void)draw:(NSPoint)point;
-(void)draw:(NSPoint)point size:(NSSize)size bounds:(NSRect)bounds;

-(NSString *)filename;
-(int)filesize;
-(NSDate *)date;
-(NSString *)format;
-(int)width;
-(int)height;
-(int)depth;
-(BOOL)transparent;
-(NSColor *)backgroundColor;

-(void)setFormat:(NSString *)fmt;
-(void)setDepth:(int)d;
-(void)setTransparent:(BOOL)trans;
-(void)setBackgroundColor:(NSColor *)col;

-(NSString *)descriptiveFilesize;
-(NSString *)descriptiveDate;
-(NSString *)descriptiveDepth;

@end



@interface XeeMultiImage:XeeImage
{
	NSMutableArray *subimages;
	int currindex;
}

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes view:(XeeView *)targetview;
-(void)dealloc;

-(int)frames;
-(void)setFrame:(int)frame;
-(int)frame;

-(void)draw:(NSPoint)point size:(NSSize)size bounds:(NSRect)bounds;

-(int)width;
-(int)height;
-(int)depth;
-(BOOL)transparent;
-(NSColor *)backgroundColor;

@end



#define XeeBitmapTypeLuma8 1
#define XeeBitmapTypeYUV422 2
#define XeeBitmapTypeRGB24 3
#define XeeBitmapTypeARGB32 4
#define XeeBitmapTypeBGRA32 5

@interface XeeBitmapImage:XeeImage
{
	int type,bytesperrow,pixelsize;
	BOOL premultiplied;

	void *data;
	BOOL freedata;
	id datasource;

	GLuint textarget,texintformat,texformat,textype;

	GWorldPtr gw;

	NSMutableArray *tiles;
}

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes view:(XeeView *)targetview;
-(id)initWithType:(int)pixeltype width:(int)framewidth height:(int)frameheight view:(XeeView *)targetview;
-(id)initWithData:(void *)pixeldata type:(int)pixeltype width:(int)framewidth height:(int)frameheight bytesPerRow:(int)bprow view:(XeeView *)targetview;

-(void)dealloc;

-(void)setData:(void *)data type:(int)type width:(int)framewidth height:(int)frameheight bytesPerRow:(int)bprow;
-(BOOL)allocBitmapWithType:(int)pixeltype width:(int)framewidth height:(int)frameheight;
-(void)setDataSource:(id)source;
-(void)setPremultiplied:(BOOL)pre;

-(void)buildTextures;
-(void)allocTexturesRect;
-(void)allocTextures2D;
-(void)loadTextures;
-(void)fixYUVGamma;
-(void)fixYUVGammaFast;
-(void)fixYUVGammaSlow;

-(void)draw:(NSPoint)point size:(NSSize)size bounds:(NSRect)bounds;

-(int)type;
-(int)bytesPerRow;
-(void *)data;
-(GWorldPtr)gworld;

@end

@interface XeeImageTile:NSObject
{
	int x,y;
	int width,height;
	float texwidth,texheight;

	GLuint tex;
}

-(id)initWithTarget:(GLuint)textarget x:(int)x0 y:(int)y0 width:(int)w height:(int)h texwidth:(float)tw texheight:(float)th;
-(void)loadWithTarget:(GLuint)textarget internalFormat:(GLuint)texintformat format:(GLuint)texformat type:(GLuint)textype data:(void *)data;
-(void)drawWithTarget:(GLuint)textarget bounds:(NSRect)bounds;

@end

