#import "XeeQuicktimeLoader.h"
#import "XeeNSImageLoader.h"



@implementation XeeQuicktimeImage

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes view:(XeeView *)targetview;
{
	if(self=[super initWithFile:name firstBlock:block attributes:attributes view:targetview])
	{
		gi=NULL;

		if([[filename pathExtension] caseInsensitiveCompare:@"bmp"]!=NSOrderedSame
		&&[attrs fileHFSTypeCode]!='BMP '&&[attrs fileHFSTypeCode]!='BMP ') // explicitly exclude BMP - what the hell, Quicktime?
		{
			CFURLRef url=CFURLCreateWithFileSystemPath(NULL,(CFStringRef)filename,kCFURLPOSIXPathStyle,FALSE);
			if(url)
			{
				FSRef fsref;
				if(CFURLGetFSRef(url,&fsref))
				{
					CFRelease(url);

					FSSpec fsspec;
					if(FSGetCatalogInfo(&fsref,kFSCatInfoNone,NULL,NULL,&fsspec,NULL)==noErr)
					{
						if(GetGraphicsImporterForFile(&fsspec,&gi)==noErr)
						{
							ImageDescriptionHandle desc;

							if(GraphicsImportGetImageDescription(gi,&desc)==noErr)
							{
								width=(*desc)->width;
								height=(*desc)->height;
								depth=(*desc)->depth;
								[self setFormat:[[[NSString alloc] initWithBytes:(*desc)->name+1 length:(*desc)->name[0] encoding:NSMacOSRomanStringEncoding] autorelease]];

								DisposeHandle((Handle)desc);

								if([format isEqual:@"Photo - JPEG"]) [self setFormat:@"JPEG"]; // what the hell, Quicktime?

								nextindex=0;

								return self;
							}
						}
					}
				}
			}
		}

		[self release];
	}

	return nil;
}

-(void)dealloc
{
	if(gi) CloseComponent(gi);

	[super dealloc];
}


static OSErr progress_func(short message,Fixed completeness,long refcon)
{
	XeeQuicktimeImage *image=(XeeQuicktimeImage *)refcon; // ow ow ow, 64-bit issues!

	if([image hasBeenStopped]) return codecAbortErr;
	return noErr;
}

-(BOOL)load
{
	stop=NO;

	EnterMoviesOnThread(0);

	CGrafPtr origport;
	GDHandle origdev;
	GetGWorld(&origport,&origdev);

	unsigned long count;
	GraphicsImportGetImageCount(gi,&count);

	struct ICMProgressProcRecord progrec={progress_func,(long)self}; // ouch, 64-bit issues!
	GraphicsImportSetProgressProc(gi,&progrec);

	GraphicsImportSetQuality(gi,codecLosslessQuality);
	while(nextindex<count)
	{
		if(stop) break;

		if(GraphicsImportSetImageIndex(gi,nextindex+1)==noErr)
		{
			ImageDescriptionHandle desc;
			if(GraphicsImportGetImageDescription(gi,&desc)==noErr)
			{
				int framewidth=(*desc)->width;
				int frameheight=(*desc)->height;
				int framedepth=(*desc)->depth;
				DisposeHandle((Handle)desc);

				int type;
				if([format isEqual:@"JPEG"]) type=XeeBitmapTypeYUV422;
				else type=XeeBitmapTypeARGB32;

				XeeBitmapImage *image=[[[XeeBitmapImage alloc] initWithType:type width:framewidth height:frameheight view:view] autorelease];

				if(image)
				{
					short drawsall;

					if(GraphicsImportDoesDrawAllPixels(gi,&drawsall)!=noErr || drawsall!=graphicsImporterDrawsAllPixels)
					{
						[image setTransparent:YES];
						[image setBackgroundColor:[NSColor whiteColor]];
					}

					[image setDepth:framedepth];

					if(GraphicsImportSetGWorld(gi,[image gworld],NULL)==noErr)
					{
						ComponentResult res=GraphicsImportDraw(gi);
						if(res==noErr)
						{
							if([format isEqual:@"JPEG"]) [image fixYUVGamma]; 

							[image buildTextures];
							[subimages addObject:image];
						}
						else
						{
							if(!stop) [subimages removeAllObjects];
							break;
						}
					}
				}
			}
		}
		nextindex++;
	}

	SetGWorld(origport,origdev);

	ExitMoviesOnThread();

	if(stop) return NO;

	CloseComponent(gi);
	gi=NULL;

	if(![subimages count]) // failed to load any frames, ask XeeNSImage for help
	{
		NSImage *image=[[[NSImage alloc] initWithContentsOfFile:filename] autorelease];
		NSArray *images=[XeeNSImage convertRepresentations:[image representations] view:view];
		if(images) [subimages addObjectsFromArray:images];

		return [subimages count]?YES:NO;
	}

	return YES;
}

-(BOOL)loaded
{
	return gi?NO:YES;
}



+(NSArray *)fileTypes
{
	return [NSArray arrayWithObjects:
		@"jpg", // JPEG Image
		@"jpeg",
		@"jpe",
		@"'JPEG'",

		@"png", // Portable Network Graphics
		@"'PNG '", 
		@"'PNGf'", 

		@"gif", // Graphics Interchange Format
		@"'GIFf'",
		@"'GIF '",

		@"tif", // TIFF Image
		@"tiff", 
		@"'TIFF'", 

		@"bmp", // Windows Bitmap Image
		@"dib", //?
		@"'BMP '", 
		@"'BMPf'", 
		@"'BMPp'", 

		@"psd", // Adobe Photoshop Image
		@"'8BPS'",

		@"tga", // Targa Image
		@"'TPIC'", 

		@"jp2", // JPEG 2000 Image
		@"'jp2 '", 

		@"pict", // PICT Image
		@"pct", 
		@"pic",
		@"'PICT'", 

		@"fpx", // FlashPix Image
		@"'FPix'", 

		@"qtif", // Apple Quicktime Image
		@"qti",
		@"qif", 
		@"'qtif'", 

		@"sgi", // Silicon Graphics Image
		@"'SGI '", 
		@"'.SGI'", 

		@"pntg", // Apple MacPaint image
		@"'PNTG'", 

		@"mac", //?

		nil
	];
}

@end
